%%%% FIR_CodeGen.m: Generate Verilog code for multiplierless FIR filters.
%Copyright (C) 2005  Douglas L Maskell

%This program is free software; you can redistribute it and/or
%modify it under the terms of the GNU General Public License
%as published by the Free Software Foundation; either version 2
%of the License, or (at your option) any later version.

%This program is distributed in the hope that it will be useful,
%but WITHOUT ANY WARRANTY; without even the implied warranty of
%MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%GNU General Public License for more details.

function []=FIR_CodeGen(CSD_Coeff, Even_flag, Direct_flag)
% The program takes the optimised coefficients and generates verilog code
% Even_flag=1 (odd No. of taps); Even_flag=0 (even No. of taps);
% Direct_flag=1 (use direct implementation); Direct_flag=0 (use min adder depth);
%       Generally you would use the minimum adder depth
% Filt_struct   = |root|L1;R1|L1L;L1R;R1L;R1R|L1LL;L1LR;L1RL;L1Rr;R1LL;R1LR;R1RL;R1RR|....
%               = | 1  | 2;3 | 4 ; 5 ; 6 ; 7 |  8 ; 9  ; 10 ; 11 ; 12 ; 13 ; 14 ; 15 | 16 ....

%%% enter CSD coefficients from other source. The optomising filter program
%%%     FIR_design_opt.m will generate these. variable is "CSD_Coeff"
%**** NOTE ONLY enter the 1st half of the coeffieients and the even flag value) ****%

% Global variables
global N;       % Filter order
global M;       % Coefficient word length
global WLen;    % Input word length
global add_depth;    % max adder depth allowed
global orig_coeff;   % Original single coefficient in CSD form
global End_Flag;   % This is 999
global dec_coeff;   % decimal value of the coefficients

if nargin~=3,
   error('USEAGE: FIR_CodeGen(CSD_Coeff, Even_flag, Direct_flag);')
end

% data enquiry
S=sprintf('\nThis is a program to generate the verilog code (FIR.v) for a digital filter.\n');
disp(S)
S=sprintf('\nCertain variables must be setup before running this program. Check code for details.\n');
disp(S)
S=sprintf('\nIgnore the screen output from this program. It is for debugging only.\n');
disp(S)

WLen=input('What is the input signal wordsize => ');    % This is generally 16 bits

% The FIR_design_opt.m generates 1 redundant MSB zero which we need to get rid of
%       and 1 zero to the left of the implied decimal point which we ignore
[N,M]=size(CSD_Coeff);	% M is the coefficient word length.
CSD_Coeff=CSD_Coeff(:,2:M)

% generate decimal values
% Calculate power of two vector
[N,M]=size(CSD_Coeff);	% M is the new coefficient word length.
power_of_2=zeros(1,M);
power_of_2(1)=1;
for i=1:M
    power_of_2(i)=2^(-i);
end
% Get floating point version of truncated coefficients 
coeff1=zeros(N,1);
for i=1:N
    dec_coeff(i)=sum(CSD_Coeff(i,:).*power_of_2);
end
dec_coeff

%now we start
%%%%% NOTE CSD_Coeff must have MSB all zeros %%%%
End_Flag=999;
F_CSD_Coeff=fliplr(CSD_Coeff);
[N,M]=size(F_CSD_Coeff);	% M is the coefficient word length.
N=N-1;		% The filter order (Taps=N+1)
T1=nnz(F_CSD_Coeff);

fid = fopen('FIR.v','w');
fprintf(fid,'// Verilog code generation for filter with %d taps and %d non-zero bits\n', 2*N+Even_flag, T1);
fprintf(fid,'// Filter has a coefficient wordlength of %d bits and a data wordsize of %d bits\n', M, WLen);

% calculate the adder depth
maxBits=0;
for I=0:N
  noBits_tmp=nnz(F_CSD_Coeff(I+1,:));
  if noBits_tmp > maxBits
    maxBits = noBits_tmp;
  end
end
% use equation to calculate adder depth
if Direct_flag == 0
   % Use min adderdepth
   add_depth=ceil(log2(maxBits))
else
   % Use direct implementation
   add_depth=maxBits-1
end

% Extract the common subexpressions
% On return, 101=6; 10-1=7; 1001=8; and 100-1=9. {6-9} can be negative.
newCSD_Coeff = CSE_extract(F_CSD_Coeff)
% generate the CSE modules
gen_CSE_Init(fid);

%Generate data structure
Filt_struct=gen_filt_struct(add_depth, newCSD_Coeff, F_CSD_Coeff);

Filt_struct
%pause


%%%***** Generate Multiplier block adders *****%%%
for I=0:N
  orig_coeff=F_CSD_Coeff(I+1,:);    % this is a global variable, need to check if used????
  % Generate the verilog modules for each multiplier and structural adder/register
  gen_CSE_code(Filt_struct, I, fid, Even_flag);
end

%%%***** stitch the modules together to form filter *****%%% 
    % write out verilog module header information
fprintf(fid,'\n\n//Top level module for filter\n');
fprintf(fid,'//This module stitches together all the submodules to form the final filter\n');
fprintf(fid,'module filter_top(Clk, Reset, Xin, Y_out);\n');
fprintf(fid,'\tinput Clk, Reset;\n');
fprintf(fid,'\tinput [%d:0] Xin;\n', WLen-1);
fprintf(fid,'\toutput [%d:0] Y_out;\n', WLen+M-1);  % May want to round Y_out back to WLen bits
%fprintf(fid,'\treg [%d:0] Y_out;\n', WLen+M-1);

fprintf(fid,'\n\treg [%d:0] TR_in;\n', WLen-1);     % this is for timing on Xin

fprintf(fid,'\n');
% Define wires to connect modules
%for I=1:4       % these are for CSEs
  % 101=1; 10-1=2; 1001=3; and 100-1=4. 
  fprintf(fid,'\twire [%d:0] CSEW_%d;\n', WLen+2, 1);  
  fprintf(fid,'\twire [%d:0] CSEW_%d;\n', WLen+1, 2);  
  fprintf(fid,'\twire [%d:0] CSEW_%d;\n', WLen+3, 3);  
  fprintf(fid,'\twire [%d:0] CSEW_%d;\n', WLen+2, 4);  
%end
for I=0:N       % these are for filter modules
  fprintf(fid,'\twire [%d:0] TW1_%d, TW2_%d;\n', WLen+M, I, I);  
end
fprintf(fid,'\twire [%d:0] TWend;\n', WLen+M);  % this is the end wire

% Now define modules
% First the CSEs
fprintf(fid,'\n');
for I=1:4
    fprintf(fid,'\tfilt_CSE_%d FCSE_%d(TR_in, CSEW_%d);\n', I, I, I);
end

% Now the filter modules
fprintf(fid,'\n');
%for I=0
  fprintf(fid,'\tfilt_0 FM_0(Clk, Reset, TR_in, CSEW_1, CSEW_2, CSEW_3, CSEW_4, TW2_1, TW1_1, TW2_0);\n');
for I=1:N
  if I == N             % Last module
     if Even_flag == 1 	% Last module and odd number of taps  
        fprintf(fid,'\tfilt_%d FM_%d(Clk, Reset, TR_in, CSEW_1, CSEW_2, CSEW_3, CSEW_4, TW1_%d, TW2_%d);\n',I,I,I,I);
     else
        fprintf(fid,'\tfilt_%d FM_%d(Clk, Reset, TR_in, CSEW_1, CSEW_2, CSEW_3, CSEW_4, TW1_%d, TWend, TWend, TW2_%d);\n',I,I,I,I);
     end
  else
    fprintf(fid,'\tfilt_%d FM_%d(Clk, Reset, TR_in, CSEW_1, CSEW_2, CSEW_3, CSEW_4, TW1_%d, TW2_%d, TW1_%d, TW2_%d);\n',I,I,I,I+1,I+1,I);
  end
end

% Now we need to fix the precision of Y_out
fprintf(fid,'\n');
fprintf(fid,'\tassign Y_out = TW2_0[%d:0];\n', WLen+M-1);
fprintf(fid,'\n');

    fprintf(fid,'\talways @(posedge Clk)\n');
    fprintf(fid,'\t\tif (Reset) TR_in <= 0;\n');             
    fprintf(fid,'\t\telse TR_in <= Xin;\n\n');			% Latch the input (Xin)

% now close module
fprintf(fid,'endmodule\n\n');

fclose(fid);
% End of programme




% Internal functions
%%******************************************************************%%
function CSE_Coeff = CSE_extract(newCSD_Coeff)
  % function extracts Common Sub Expressions
  % On return, 101=6; 10-1=7; 1001=8; and 100-1=9. {6-9} can be negative.
  % ie -101 will be -7
  
  CSE_Coeff = newCSD_Coeff;
  [N,M]=size(newCSD_Coeff);
  for I=1:N
    CSD_Coeff=newCSD_Coeff(I,:);  
    for K=1:M
      if CSD_Coeff(K) ~= 0
        if K == M-1
            break           % this catches end of coefficient case.
        elseif CSD_Coeff(K+2) ~= 0
          % this is 101 or 10-1
          if CSD_Coeff(K) == 1
            % 101 or 10-1   (ie 6 or 7)
            if CSD_Coeff(K+2) == 1  %101
              CSD_Coeff(K) = 6;
              CSD_Coeff(K+2) = 0;
            else                    %10-1
              CSD_Coeff(K) = 7;
              CSD_Coeff(K+2) = 0;
            end    
          else
            % -10-1 or -101 (ie -6 or -7) 
            if CSD_Coeff(K+2) == 1  %-101
              CSD_Coeff(K) = -7;
              CSD_Coeff(K+2) = 0;
            else                    %-10-1
              CSD_Coeff(K) = -6;
              CSD_Coeff(K+2) = 0;
            end    
          end
        elseif K == M-2
            break           % this catches end of coefficient case.
        elseif CSD_Coeff(K+3) ~= 0
          % this is 1001 or 100-1
          if CSD_Coeff(K) == 1
            % 1001 or 100-1   (ie 8 or 9)
            if CSD_Coeff(K+3) == 1  %1001
              CSD_Coeff(K) = 8;
              CSD_Coeff(K+3) = 0;
            else                    %100-1
              CSD_Coeff(K) = 9;
              CSD_Coeff(K+3) = 0;
            end    
          else
            % -100-1 or -1001 (ie -8 or -9) 
            if CSD_Coeff(K+2) == 1  %-1001
              CSD_Coeff(K) = -9;
              CSD_Coeff(K+3) = 0;
            else                    %-100-1
              CSD_Coeff(K) = -8;
              CSD_Coeff(K+3) = 0;
            end    
          end          
        end
      end
    end
    CSE_Coeff(I,:) = CSD_Coeff;
  end
  
  return
%%****** End of function CSE_extract *******%%



%%******************************************************************%%
function [] = gen_CSE_Init(fid)
  % function doesnot return any value. It just writes to file

% Global variables
global N;       % Filter order
global M;       % Coefficient word length
global WLen;    % Input word length
global add_depth;    % max adder depth allowed
global orig_coeff;   % Original single coefficient in CSD form


  % need to print the modules for the 4 common subexpressions {101; 10-1; 1001; 100-1}
  fprintf(fid,'// CSEs are {101, 10-1, 1001, and 100-1}\n\n', M, WLen);
  for K=1:4     % 4 CSEs
    % write out verilog module header information
    fprintf(fid,'module filt_CSE_%d(Xin, CSE_O_%d);\n', K, K);
    fprintf(fid,'\tinput [%d:0] Xin;\n', WLen-1);
    if K == 1   % 101 case
      fprintf(fid,'\toutput [%d:0] CSE_O_%d;\n', WLen+2, K);
      fprintf(fid,'\tassign CSE_O_%d = {{{{3{Xin[%d]}}, Xin[%d:2]} + {{1{Xin[%d]}}, Xin[%d:0]}}, Xin[1:0]};\n', K, WLen-1, WLen-1, WLen-1, WLen-1);
    elseif K==2 % 10-1 case
      fprintf(fid,'\toutput [%d:0] CSE_O_%d;\n', WLen+1, K);
      fprintf(fid,'\tassign CSE_O_%d = {{{{2{Xin[%d]}}, Xin[%d:2]} - Xin[%d:0]}, Xin[1:0]};\n', K, WLen-1, WLen-1, WLen-1);
    elseif K==3 % 1001 case
      fprintf(fid,'\toutput [%d:0] CSE_O_%d;\n', WLen+3, K);
      fprintf(fid,'\tassign CSE_O_%d = {{{{3{Xin[%d]}}, Xin[%d:3]} + {{1{Xin[%d]}}, Xin[%d:0]}}, Xin[2:0]};\n', K, WLen-1, WLen-1, WLen-1, WLen-1);
    else        %100-1 case
      fprintf(fid,'\toutput [%d:0] CSE_O_%d;\n', WLen+2, K);
      fprintf(fid,'\tassign CSE_O_%d = {{{{2{Xin[%d]}}, Xin[%d:3]} - Xin[%d:0]}, Xin[2:0]};\n', K, WLen-1, WLen-1, WLen-1);
    end
    fprintf(fid,'endmodule\n\n');
  end
  return
%%****** End of function gen_CSE_Init *******%%




%%******************************************************************%%
function Filt_struct=gen_filt_struct(add_depth, newCSD_Coeff, orig_CSD_Coeff)
  % function sets up the Binary heap data structure.
  % Note: CSD_Coeff 101=6; 10-1=7; 1001=8; and 100-1=9. 
  % {6-9} can be negative. ie -101 will be -7
  % Full is 999 (End_Flag)


%% Global variables
%global N;       % Filter order (Taps=N+1)
%global M;       % Coefficient word length
%global WLen;    % Input word length
%global add_depth;    % max adder depth allowed
%global orig_coeff;   % Original single coefficient in CSD form
global End_Flag;   % This is 999

T_lev = add_depth+1;
[N,M]=size(newCSD_Coeff);
Filt_struct=zeros(2*N,(2^T_lev)-1);
  
for L=0:N-1  
  spare_bits=2^add_depth-nnz(orig_CSD_Coeff(L+1,:));
  CSD_Coeff=newCSD_Coeff(L+1,:);  

  i_nz_Bits=find(CSD_Coeff);
  [x,y]=size(i_nz_Bits);
  if y == 0
     %This coefficient is zero, data structure will be zero
     % do nothing
  elseif y == 1
     % This coefficient has a single bit or a single CSE term
     Filt_struct(2*L+2,1)=i_nz_Bits(y);
     Filt_struct(2*L+1,1)=CSD_Coeff(i_nz_Bits(y));
  else
     % Multiple terms
     % initially just fill array
     A_index=(2^(add_depth-1))-1;
     B_index=(2^add_depth)-1;     
     for I=1:A_index
        Filt_struct(2*L+1,I)=End_Flag;
     end
     ones_Full_flag=0;
     fix_Flag=0;
     for I=1:y
        I_Ptr=I-fix_Flag;
        if abs(CSD_Coeff(i_nz_Bits(I))) == 1
           % up to next level
           ones_Full_flag=ones_Full_flag+1;
           if 2*fix(ones_Full_flag/2) ~= ones_Full_flag
              % have new entry
              Filt_struct(2*L+1,A_index+I_Ptr)=End_Flag;
              Filt_struct(2*L+1,2*(A_index+I_Ptr))=CSD_Coeff(i_nz_Bits(I));
              Filt_struct(2*L+2,2*(A_index+I_Ptr))=i_nz_Bits(I);
              save_index=I;
           else
              % check if we have any slack, if so add a seperate entry
              if I == save_index+1
                 % consecquitave entries and already have an entry, so find it
                 for K=(2^T_lev)-1:-1:2^(add_depth)
                    if abs(Filt_struct(2*L+1,K)) == 1
                       Filt_struct(2*L+1,K+1)=CSD_Coeff(i_nz_Bits(I));
                       Filt_struct(2*L+2,K+1)=i_nz_Bits(I);
                       fix_Flag=fix_Flag+1;
                       break
                    end
                 end
              elseif spare_bits > 0;
                 % Can have new entry
                 Filt_struct(2*L+1,A_index+I_Ptr)=End_Flag;
                 Filt_struct(2*L+1,2*(A_index+I_Ptr))=CSD_Coeff(i_nz_Bits(I));
                 Filt_struct(2*L+2,2*(A_index+I_Ptr))=i_nz_Bits(I);
                 save_index=I;
                 spare_bits=spare_bits-1;
              else
                 % No room so stick it with existing entry 
                 for K=(2^T_lev)-1:-1:2^(add_depth)
                    if abs(Filt_struct(2*L+1,K)) == 1
                       Filt_struct(2*L+1,K+1)=CSD_Coeff(i_nz_Bits(I));
                       Filt_struct(2*L+2,K+1)=i_nz_Bits(I);
                       fix_Flag=fix_Flag+1;
                    end
                 end
              end
           end
        else
           Filt_struct(2*L+1,A_index+I_Ptr)=CSD_Coeff(i_nz_Bits(I));
           Filt_struct(2*L+2,A_index+I_Ptr)=i_nz_Bits(I);
        end
     end
  end
  
  %continue
  
  % now need to relax the structure
  % Firstly remove single entry nodes
  for I=(2^T_lev)-1:-1:1
     % find last full (999) node
     if Filt_struct(2*L+1,I) == End_Flag
        % check for empty nodes connected to this one.
        if Filt_struct(2*L+1,2*I) == 0
           % this branch has zero structure
           Filt_struct(2*L+1,I)= 0;
           Filt_struct(2*L+2,I)= 0;
        end
        if Filt_struct(2*L+1,2*I+1) == 0
           if Filt_struct(2*L+1,2*I) == End_Flag
              % Ignore this as it has a tree structure above it.
           else
              Filt_struct(2*L+1,I)= Filt_struct(2*L+1,2*I);
              Filt_struct(2*L+1,2*I)=0;
              Filt_struct(2*L+2,I)= Filt_struct(2*L+2,2*I);
              Filt_struct(2*L+2,2*I)=0;
           end
        end
     end
  end
  
  %continue
  
  % Now remove unwanted nodes
  done = false;
  while ~done     % may need a number of iterations to fix structure.
     found_flag=0;
     for I=1:(2^T_lev)-1;
       % find full (999) nodes
       if Filt_struct(2*L+1,I) == End_Flag
          % check for empty nodes connected to this one.
          % Note, should only be the RH node which can be empty.
          if Filt_struct(2*L+1,2*I) == 0
             % this branch has zero structure, ERROR
             disp(' Error in filter structure')
             L
             Filt_struct(2*L+1,:)
             Filt_struct(2*L+2,:)
             pause
          end
          if Filt_struct(2*L+1,2*I+1) == 0    % have a Null RH branch.
             % first check if we can remove tree structure completely
             % check if RHS of LHB is full, if so juggle down
             if (I < (2^(T_lev-3)) & Filt_struct(2*L+1,2*(2*I)+1) == End_Flag)
                % Juggle RHS structure.
                found_flag=1;
                IA_Dest=2*2*I+1;
                IA_Src=2*IA_Dest;
                IB_Dest=2*I+1;
                IB_Src=IA_Src+1;
                J_ptr=1;
                done1 = false;
                while ~done1 
                   if IA_Src > (2^(T_lev))-1   % index will be outside index of Filt_struct
                      break
                   end
                   for K1=0:J_ptr-1
                      Filt_struct(2*L+1,IA_Dest+K1)= Filt_struct(2*L+1,IA_Src+K1);
                      Filt_struct(2*L+2,IA_Dest+K1)= Filt_struct(2*L+2,IA_Src+K1);
                      Filt_struct(2*L+1,IB_Dest+K1)= Filt_struct(2*L+1,IB_Src+K1);
                      Filt_struct(2*L+2,IB_Dest+K1)= Filt_struct(2*L+2,IB_Src+K1);
                      Filt_struct(2*L+1,IA_Src+K1)=0;
                      Filt_struct(2*L+2,IA_Src+K1)=0;
                      Filt_struct(2*L+1,IB_Src+K1)=0;
                      Filt_struct(2*L+2,IB_Src+K1)=0;
                   end
                   IA_Dest=2*IA_Dest;
                   IA_Src=2*IA_Src;
                   IB_Dest=2*IB_Dest;
                   IB_Src=2*IB_Src;
                   J_ptr=2*J_ptr;
                end
                break    % break out of outside for loop
             else % Just Elliminate RHS structure.
                found_flag=1;
                I_ptr=I;
                J_ptr=2;
                done1 = false;
                while ~done1 
                   if I_ptr > (2^(T_lev-2))-1   % index will be outside index of Filt_struct
                      break
                   end
                   for K1=0:J_ptr-1
                      Filt_struct(2*L+1,2*I_ptr+K1)= Filt_struct(2*L+1,2*(2*I_ptr)+K1);
                      Filt_struct(2*L+2,2*I_ptr+K1)= Filt_struct(2*L+2,2*(2*I_ptr)+K1);
                      Filt_struct(2*L+1,2*(2*I_ptr)+K1)= 0;
                      Filt_struct(2*L+2,2*(2*I_ptr)+K1)= 0;
                   end
                   I_ptr=2*I_ptr;
                   J_ptr=2*J_ptr;
                end
                break    % break out of outside for loop
             end
          end
       end
     end
     if found_flag ~= 1
        break
     end
  end
end
return
%%****** End of function gen_filt_struct *******%%




%%******************************************************************%%
function [] = gen_CSE_code(F_struct, I, fid, Even_flag)
  % function doesnot return any value. It just writes to file

% Global variables
global N;       % Filter order
global M;       % Coefficient word length
global WLen;    % Input word length
global add_depth;    % max adder depth allowed
global orig_coeff;   % Original single coefficient in CSD form
global End_Flag;   % This is 999
global dec_coeff;   % decimal value of the coefficients


  if I == 0     % first module 
    % write out verilog module header information
    fprintf(fid,'module filt_%d(Clk, Reset, Xin, CSE_I_1, CSE_I_2, CSE_I_3, CSE_I_4, In2_%d, Out1_%d, Out2_%d);\n',I,I,I,I);
    fprintf(fid,'// Value of filter coefficient is: %18.16f\n', dec_coeff(I+1));
    fprintf(fid,'\tinput Clk, Reset;\n');
    fprintf(fid,'\tinput [%d:0] Xin;\n', WLen-1);
    fprintf(fid,'\tinput [%d:0] CSE_I_1;\n', WLen+2);
    fprintf(fid,'\tinput [%d:0] CSE_I_2;\n', WLen+1);
    fprintf(fid,'\tinput [%d:0] CSE_I_3;\n', WLen+3);
    fprintf(fid,'\tinput [%d:0] CSE_I_4;\n', WLen+2);
    fprintf(fid,'\tinput [%d:0] In2_%d;\n', WLen+M, I);
    fprintf(fid,'\toutput [%d:0] Out1_%d, Out2_%d;\n', WLen+M, I, I);
    fprintf(fid,'\treg [%d:0] Out1_%d, Out2_%d;\n', WLen+M, I, I);
    C_sign=Gen_coeff_paper(F_struct, I, fid);	% This generates the coefficient
    fprintf(fid,'\talways @(posedge Clk)\n');
    if C_sign == 0
      fprintf(fid,'\t\tif (Reset) begin\n\t\t\tOut1_%d <= 0;\n\t\t\tOut2_%d <= 0;\n\t\t\tend\n',I,I);		% Mult_out is the output of the current MB
      fprintf(fid,'\t\telse begin\n\t\t\tOut1_%d <= Mult_out;\n',I);
      fprintf(fid,'\t\t\tOut2_%d <= In2_%d + Mult_out;\n\t\t\tend\n',I, I);		% Mult_out is the output of the current MB
    else
      fprintf(fid,'\t\tif (Reset) begin\n\t\t\tOut1_%d <= 0;\n\t\t\tOut2_%d <= 0;\n\t\t\tend\n',I,I);		% Mult_out is the output of the current MB
      fprintf(fid,'\t\telse begin\n\t\t\tOut1_%d <= -Mult_out;\n',I);
      fprintf(fid,'\t\t\tOut2_%d <= In2_%d - Mult_out;\n\t\t\tend\n',I, I);		% Mult_out is the output of the current MB
    end
    fprintf(fid,'endmodule\n\n');
    
  elseif (I == N & Even_flag == 1) 	% Last module and odd number of taps
    % write out verilog module header information
    fprintf(fid,'module filt_%d(Clk, Reset, Xin, CSE_I_1, CSE_I_2, CSE_I_3, CSE_I_4, In1_%d, Out2_%d);\n',I,I,I);
    fprintf(fid,'// Value of filter coefficient is: %18.16f\n', dec_coeff(I+1));
    fprintf(fid,'\tinput Clk, Reset;\n');
    fprintf(fid,'\tinput [%d:0] Xin;\n', WLen-1);
    fprintf(fid,'\tinput [%d:0] CSE_I_1;\n', WLen+2);
    fprintf(fid,'\tinput [%d:0] CSE_I_2;\n', WLen+1);
    fprintf(fid,'\tinput [%d:0] CSE_I_3;\n', WLen+3);
    fprintf(fid,'\tinput [%d:0] CSE_I_4;\n', WLen+2);
    fprintf(fid,'\tinput [%d:0] In1_%d;\n', WLen+M, I);
    fprintf(fid,'\toutput [%d:0] Out2_%d;\n', WLen+M, I);
    fprintf(fid,'\treg [%d:0] Out2_%d;\n', WLen+M, I);
    C_sign=Gen_coeff_paper(F_struct, I, fid);	% This generates the coefficient
    fprintf(fid,'\talways @(posedge Clk)\n');
    if C_sign == 0
      fprintf(fid,'\t\tif (Reset) Out2_%d <= 0;\n',I);		% Mult_out is the output of the current MB
      fprintf(fid,'\t\telse Out2_%d <= In1_%d + Mult_out;\n',I, I);	% Mult_out is the output of the current MB
    else
      fprintf(fid,'\t\tif (Reset) Out2_%d <= 0;\n',I);		% Mult_out is the output of the current MB
      fprintf(fid,'\t\telse Out2_%d <= In1_%d - Mult_out;\n',I, I);	% Mult_out is the output of the current MB
    end
    fprintf(fid,'endmodule\n\n');
    
  else		% Middle module/End module(even taps)
    % write out verilog module header information
    fprintf(fid,'module filt_%d(Clk, Reset, Xin, CSE_I_1, CSE_I_2, CSE_I_3, CSE_I_4, In1_%d, In2_%d, Out1_%d, Out2_%d);\n',I,I,I,I,I);
    fprintf(fid,'// Value of filter coefficient is: %18.16f\n', dec_coeff(I+1));
    fprintf(fid,'\tinput Clk, Reset;\n');
    fprintf(fid,'\tinput [%d:0] Xin;\n', WLen-1);
    fprintf(fid,'\tinput [%d:0] CSE_I_1;\n', WLen+2);
    fprintf(fid,'\tinput [%d:0] CSE_I_2;\n', WLen+1);
    fprintf(fid,'\tinput [%d:0] CSE_I_3;\n', WLen+3);
    fprintf(fid,'\tinput [%d:0] CSE_I_4;\n', WLen+2);
    fprintf(fid,'\tinput [%d:0] In1_%d, In2_%d;\n', WLen+M, I, I);
    fprintf(fid,'\toutput [%d:0] Out1_%d, Out2_%d;\n', WLen+M, I, I);
    fprintf(fid,'\treg [%d:0] Out1_%d, Out2_%d;\n', WLen+M, I, I);
    C_sign=Gen_coeff_paper(F_struct, I, fid);	% This generates the coefficient
    fprintf(fid,'\talways @(posedge Clk)\n');
    if C_sign == 0
      fprintf(fid,'\t\tif (Reset) begin\n\t\t\tOut1_%d <= 0;\n\t\t\tOut2_%d <= 0;\n\t\t\tend\n',I,I);		% Mult_out is the output of the current MB
      fprintf(fid,'\t\telse begin\n\t\t\tOut1_%d <= In1_%d + Mult_out;\n',I, I);
      fprintf(fid,'\t\t\tOut2_%d <= In2_%d + Mult_out;\n\t\t\tend\n',I, I);		% Mult_out is the output of the current MB
    else
      fprintf(fid,'\t\tif (Reset) begin\n\t\t\tOut1_%d <= 0;\n\t\t\tOut2_%d <= 0;\n\t\t\tend\n',I,I);		% Mult_out is the output of the current MB
      fprintf(fid,'\t\telse begin\n\t\t\tOut1_%d <= In1_%d - Mult_out;\n',I, I);
      fprintf(fid,'\t\t\tOut2_%d <= In2_%d - Mult_out;\n\t\t\tend\n',I, I);		% Mult_out is the output of the current MB
    end
    fprintf(fid,'endmodule\n\n');
  end        
  return
%%****** End of function gen_CSE_code *******%%

  


%%******************************************************************%%
function [C_sign] = Gen_coeff_paper(F_struct, I, fid)	
% function returns Sign. It also writes to file

% Global variables
global N;       % Filter order
global M;       % Coefficient word length
global WLen;    % Input word length
global add_depth;    % max adder depth allowed
global orig_coeff;   % Original single coefficient in CSD form % check if this is used !!!
global End_Flag;   % This is 999

  %i_nz_Bits=find(newCSD_Coeff1);
  %[x,y]=size(i_nz_Bits);
  
  T_Flag=800;
  
  % print wire for Mult_out
  fprintf(fid,'\n');
  fprintf(fid,'\twire [%d:0] Mult_out;\n', WLen+M);	% Mult_out is the current MB

  % Note X_1=newCSD_Coeff1(i_nz_Bits)
  % Note X_2=i_nz_Bits
  X_1=F_struct(2*I+1,1);
  X_2=F_struct(2*I+2,1);
  if X_1 == 0;
     %This coefficient is zero, set output to zero
	 fprintf(fid,'\n');
	 fprintf(fid,'\tassign Mult_out = %d''b0;\n\n', WLen+M+1);	% Mult_out is zero
     C_sign=0;
  elseif X_1 ~= End_Flag;
     % This coefficient has a single bit or a single CSE term
     % Set output to shifted input or shifted CSE
	 fprintf(fid,'\n');
     % Note we are assuming Mult_out is WLen+M bits wide
     if X_2 == 1
        str1=sprintf('');
     else
        str1=sprintf(', %d''b0', X_2-1);
     end
     % get CSE number string
     str2=sprintf('%d', abs(X_1)-5); % subtracting 5 converts back to CSE1-CSE4
     if abs(X_1) == 1       % not a CSE  
	    fprintf(fid,'\tassign Mult_out = {{%d{Xin[%d]}}, Xin[%d:0]%s};\n\n', M-X_2+2, WLen-1, WLen-1, str1);
     else    % Must be a CSE
        % CSE inputs are CSE_I_1, CSE_I_2, CSE_I_3, CSE_I_4  
        T1case=abs(X_1);
        switch T1case
           case 6             %CSE1 {101}
	          fprintf(fid,'\tassign Mult_out = {{%d{CSE_I_%s[%d]}}, CSE_I_%s[%d:0]%s};\n\n', M-X_2-1, str2, WLen+2, str2, WLen+2, str1);
           case 7             %CSE2 {10-1}
	          fprintf(fid,'\tassign Mult_out = {{%d{CSE_I_%s[%d]}}, CSE_I_%s[%d:0]%s};\n\n', M-X_2-1, str2, WLen+1, str2, WLen+1, str1);
           case 8             %CSE3 {1001}
	          fprintf(fid,'\tassign Mult_out = {{%d{CSE_I_%s[%d]}}, CSE_I_%s[%d:0]%s};\n\n', M-X_2-2, str2, WLen+3, str2, WLen+3, str1);
           case 9             %CSE4 {100-1}
	          fprintf(fid,'\tassign Mult_out = {{%d{CSE_I_%s[%d]}}, CSE_I_%s[%d:0]%s};\n\n', M-X_2-2, str2, WLen+2, str2, WLen+2, str1);
           otherwise  % is a previously generated added
              disp('error in CSE');
        end
        %if abs(X_1) < 8     % CSE1 or CSE2
	    %   fprintf(fid,'\tassign Mult_out = {{%d{CSE_I_%s[%d]}}, CSE_I_%s[%d:0]%s};\n\n', M-X_2-1, str2, WLen+2, str2, WLen+2, str1);
        %else
	    %   fprintf(fid,'\tassign Mult_out = {{%d{CSE_I_%s[%d]}}, CSE_I_%s[%d:0]%s};\n\n', M-X_2-2, str2, WLen+3, str2, WLen+3, str1);
        %end
     end    

     if X_1 < 0     % must be negative
        C_sign = 1;
     else
        C_sign = 0;
     end
  else
     done = false;
     [M1,N1]=size(F_struct(2*I+1,:));
     TC_sign=1; % positive
     W_index=0;
     Keep_parm=zeros(2^add_depth,1);   % Keep_parm(i)=size,
     % find first adder to get location (will be in LHS of tree)
     done=false;
     K=1;
     while ~done
        if F_struct(2*I+1,2*K) == End_Flag
           K=2*K;
        else
           keep_offset=F_struct(2*I+2,2*K);   %get location of first adder for later
           done=true;
        end
     end
     
     %while ~done     % may need a number of iterations to realise the structure.
        %Root_flag=0;
        TC_sign=1; % positive
        for K=N1:-1:1
           if F_struct(2*I+1,K) == End_Flag     % found first adder
              W_index=W_index+1;
              %T_a=F_struct(2*I+2,2*K);
              %T_b=F_struct(2*I+2,2*K+1);
              %Dlen= T_b-T_a;
              % check sign of LH bit
              % If negative, make both positive and carry sign
              if sign(F_struct(2*I+1,2*K)) ~= 1   % Negative value
                 TC_sign=-1;
                 F_struct(2*I+1,2*K)=-F_struct(2*I+1,2*K);
                 F_struct(2*I+1,2*K+1)=-F_struct(2*I+1,2*K+1);
              end
              %LHS=[size, location]
              LHS_val=F_struct(2*I+1,2*K);
              locat=F_struct(2*I+2,2*K);
              T1case=abs(LHS_val);
              switch T1case
                 case 1
                    Bsize=WLen;     %Note, LHS is only Wlen
                    LHS_str=sprintf('Xin');
                 case 6
                    Bsize=WLen+1+2;
                    LHS_str=sprintf('CSE_I_1');
                 case 7
                    Bsize=WLen+1+1;
                    LHS_str=sprintf('CSE_I_2');
                 case 8
                    Bsize=WLen+1+3;
                    LHS_str=sprintf('CSE_I_3');
                 case 9
                    Bsize=WLen+1+2;
                    LHS_str=sprintf('CSE_I_4');
                 otherwise  % is a previously generated added
                    K1=abs(LHS_val)-T_Flag; 
                    Bsize=Keep_parm(K1,1)+1;
                    LHS_str=sprintf('T_%d',K1);
              end
              LHS=[Bsize, locat];
              RHS_val=F_struct(2*I+1,2*K+1);
              locat=F_struct(2*I+2,2*K+1);
              T2case=abs(RHS_val);
              switch T2case
                 case 1
                    Bsize=WLen;
                    RHS_str=sprintf('Xin');
                 case 6
                    Bsize=WLen+1+2;
                    RHS_str=sprintf('CSE_I_1');
                 case 7
                    Bsize=WLen+1+1;
                    RHS_str=sprintf('CSE_I_2');
                 case 8
                    Bsize=WLen+1+3;
                    RHS_str=sprintf('CSE_I_3');
                 case 9
                    Bsize=WLen+1+2;
                    RHS_str=sprintf('CSE_I_4');
                 otherwise  % is a previously generated added
                    K1=abs(RHS_val)-T_Flag; 
                    Bsize=Keep_parm(K1,1)+1;
                    RHS_str=sprintf('T_%d',K1);
              end
              RHS=[Bsize, locat];
              Dlen=RHS(2)-LHS(2);
              if abs(RHS_val) == 1   % need to add a sign bit to RHS term
                 %fprintf(fid,'\n\twire [%d:0] T_%d;\n', RHS(1)+Dlen, W_index);		% gen temp wires
                 fprintf(fid,'\n\twire [%d:0] T_%d;\n', RHS(1)+Dlen, W_index);		% gen temp wires
              else
                 %fprintf(fid,'\n\twire [%d:0] T_%d;\n', RHS(1)+Dlen, W_index);		% gen temp wires
                 fprintf(fid,'\n\twire [%d:0] T_%d;\n', RHS(1)+Dlen-1, W_index);		% gen temp wires
              end
              
           if LHS(2)-1 == 0
              strL=sprintf('');
           else
           %   strL=sprintf(',%d''b0', LHS(2)-1);
              strL=sprintf('');
           end
           if F_struct(2*I+1,2*K+1) < 0     % is negative
              str3=sprintf('-');
           else
              str3=sprintf('+');
           end
           Slen=RHS(1)-LHS(1)+Dlen;
           if abs(RHS_val) == 1   % need to add a sign bit to RHS term
              fprintf(fid,'\tassign T_%d = {{{{%d{%s[%d]}},%s[%d:%d]}%s{{1{%s[%d]}},%s[%d:0]}},%s[%d:0]%s};\n',W_index,Slen+1,LHS_str,LHS(1)-1,LHS_str,LHS(1)-1,Dlen,str3,RHS_str,RHS(1)-1,RHS_str,RHS(1)-1,LHS_str,Dlen-1,strL);
              Keep_parm(W_index)=RHS(1)+Dlen;   
              %Keep_parm(W_index)=RHS(1)+RHS(2)-keep_offset;   
           else
              fprintf(fid,'\tassign T_%d = {{{{%d{%s[%d]}},%s[%d:%d]}%s{%s[%d:0]}},%s[%d:0]%s};\n',W_index,Slen,LHS_str,LHS(1)-1,LHS_str,LHS(1)-1,Dlen,str3,RHS_str,RHS(1)-1,LHS_str,Dlen-1,strL);
              Keep_parm(W_index)=RHS(1)+Dlen-1;   
              %Keep_parm(W_index)=RHS(1)+RHS(2)-keep_offset-1;   
           end
           F_struct(2*I+1,K)=TC_sign*(T_Flag+W_index);
           F_struct(2*I+2,K)=F_struct(2*I+2,2*K);
           W_index;

           end
           if K == 1    % are at the root
              if F_struct(2*I+1,K) < 0     % must be negative
                 C_sign = 1;
              else
                 C_sign = 0;
              end
              T_LHS=F_struct(2*I+2,K);
              if keep_offset-1 == 0
                 strL=sprintf('');
              else
                 strL=sprintf(', %d''b0', keep_offset-1);
              end
              K1=abs(F_struct(2*I+1,K))-T_Flag; 
              Bsize=Keep_parm(K1,1);
              str1=sprintf('T_%d',K1);
              % Mult out is size WLen+M+1
              T_size=WLen+M+1-(Bsize+1+keep_offset-1);
              fprintf(fid,'\tassign Mult_out = {{%d{%s[%d]}},%s[%d:0]%s};\n\n',T_size,str1,Bsize,str1,Bsize,strL);
           end
           %done=true;
        end
     %end
  end

  return
%%****** End of function Gen_coeff_paper *******%%



%%******************************************************************%%
function [C_sign] = implement_CSE(newCSD_Coeff1, fid, T_CSE_index, T_CSE_sign)	
% function returns Sign. It also writes to file

%now the hard bit.
     CSE_Bits=find(abs(newCSD_Coeff1) >= 6);
     [a,b]=size(CSE_Bits);      % b is the number of CSE entries in the current coefficient (row).
     i_Bits=find(abs(newCSD_Coeff1) > 0 & abs(newCSD_Coeff1) < 2);
     [a,c]=size(i_Bits);        % c is the number of single entries in the current coefficient (row).
     i1_Bits=find(abs(newCSD_Coeff1) > 2 & abs(newCSD_Coeff1) < 6);
     [a,d]=size(i1_Bits);       % d is the number of combined single entries in the current coefficient (row).
     %d=d/2;                     % d has double entries
     CSE_Bits1=find(abs(newCSD_Coeff1));
  
  iter_flag=0   
%  for I=2:b+c+d
%     % use indicies K, L, M
%     K=0; L=0; M=0;
%     if iter_flag == 1
%         iter_flag=0;
%         break
%     end
%     switch abs(newCSD_Coeff1(CSE_Bits1(I-1)))
%         case 1             % starts with a 1 or a -1
             
%         case (6 | 7)
             
 %        case (8 | 9)
             
 %        otherwise
 %           iter_flag =1;
 %           I=I+1;
 %    end
     

         
         
         
         
         

C_sign=0;

return
%%****** End of function implement_CSE *******%%

% End of functions
